@@ -212,6 +212,9 @@ def softclamp(t, value = 50.):
 def max_neg_value(t):
     return -torch.finfo(t.dtype).max
 
+def append_dims(t, ndims):
+    return t.reshape(*t.shape, *((1,) * ndims))
+
 def log(t, eps = 1e-20):
     return torch.log(t.clamp(min = eps))
 
@@ -1004,6 +1007,7 @@ class Transfusion(Module):
         to_modality_shape_fn: Callable | tuple[Callable, ...] = default_to_modality_shape_fn,
         ignore_index = -1,
         flow_loss_weight = 1.,
+        text_loss_weight = 1.,
         add_flow_direction_loss = False, # figure 7 https://arxiv.org/abs/2410.10356
         direction_loss_weight = 1.,
         velocity_consistency_loss_weight = 1.,
@@ -1153,6 +1157,7 @@ class Transfusion(Module):
 
         self.ignore_index = ignore_index
         self.flow_loss_weight = flow_loss_weight
+        self.text_loss_weight = text_loss_weight
 
         # velocity consistency weight - only added if EMA model is passed in during training
 
@@ -1506,7 +1511,7 @@ class Transfusion(Module):
         modality_type: int | None = None,
         encode_modality: bool = True,
         return_loss = True
-    ) -> Float['']:
+    ) -> Float[''] | Float['b ...']:
 
         modalities = modalities.to(self.device)
 
@@ -1535,6 +1540,8 @@ class Transfusion(Module):
 
         shape = modalities.shape
 
+        tokens = modalities
+
         # axial positions
 
         if add_pos_emb:
@@ -1547,31 +1554,21 @@ class Transfusion(Module):
 
             assert len(axial_dims) == modality_num_dim, f'received modalities of ndim {len(axial_dims)} but expected {modality_num_dim}'
 
-        # maybe transform
+        # shapes and device
 
-        tokens = transform(modalities)
+        batch, device = tokens.shape[0], tokens.device
 
         # maybe channel first
 
         if self.channel_first_latent:
             tokens = rearrange(tokens, 'b d ... -> b ... d')
 
-        tokens = rearrange(tokens, 'b ... d -> b (...) d')
-
-        # rotary
-
-        batch, seq_len, device = tokens.shape[0], tokens.shape[-2], tokens.device
-
-        pos = torch.arange(seq_len, device = device)
-
-        rotary_emb = self.rotary_emb(pos)
-
         # times
 
         if not exists(times):
             times = torch.rand((batch,), device = device)
 
-        padded_times = rearrange(times, 'b -> b 1 1')
+        padded_times = append_dims(times, tokens.ndim - 1)
 
         noise = torch.randn_like(tokens)
 
@@ -1579,8 +1576,18 @@ class Transfusion(Module):
 
         flow = tokens - noise
 
+        # from latent to model tokens
+
         noised_tokens = latent_to_model_fn(noised_tokens)
 
+        # maybe transform
+
+        noised_tokens = transform(noised_tokens)
+
+        noised_tokens = rearrange(noised_tokens, 'b ... d -> b (...) d')
+
+        seq_len = noised_tokens.shape[-2]
+
         # maybe add axial pos emb
 
         if add_pos_emb:
@@ -1592,7 +1599,6 @@ class Transfusion(Module):
         embed = self.transformer(
             noised_tokens,
             times = times,
-            rotary_emb = rotary_emb,
             modality_only = True,
         )
 
@@ -1603,6 +1609,8 @@ class Transfusion(Module):
 
         # flow loss
 
+        flow = rearrange(flow, 'b ... d -> b (...) d')
+
         return F.mse_loss(pred_flow, flow)
 
     @torch.no_grad()
@@ -1838,33 +1846,39 @@ class Transfusion(Module):
 
                 text_tensor = torch.full((modality_length,), -1, device = device) # text is all -1 here, so text labels are not learned on
 
-                # add the [som] and [eom] tokens for the modality type
+                # only add modality meta information when not returning embedding, which only occurs when sampling modality
 
-                som_id, eom_id = self.som_ids[modality_type], self.eom_ids[modality_type]
+                succeed_modality_tokens = precede_modality_tokens = 0
 
-                # start by just storing the token length of the modality
+                if not return_embed:
+                    # add the [som] and [eom] tokens for the modality type
 
-                modality_shape_str = join([*map(str, modality_shape_tuple)], ',')
-                modality_meta_info = self.char_tokenizer(modality_shape_str, device = device)
+                    som_id, eom_id = self.som_ids[modality_type], self.eom_ids[modality_type]
 
-                prec_meta_tag_length = len(modality_meta_info) + 2
+                    # start by just storing the token length of the modality
 
-                text_tensor = torch.cat((
-                    tensor_([self.meta_id]),
-                    modality_meta_info,
-                    tensor_([som_id]),
-                    text_tensor,
-                    tensor_([eom_id])
-                ))
+                    modality_shape_str = join([*map(str, modality_shape_tuple)], ',')
+                    modality_meta_info = self.char_tokenizer(modality_shape_str, device = device)
+
+                    precede_modality_tokens = len(modality_meta_info) + 2
+                    succeed_modality_tokens = 1
+
+                    text_tensor = torch.cat((
+                        tensor_([self.meta_id]),
+                        modality_meta_info,
+                        tensor_([som_id]),
+                        text_tensor,
+                        tensor_([eom_id])
+                    ))
+
+                batch_modality_positions.append((modality_type, offset + precede_modality_tokens, modality_length)) # offset + preceding meta tag length (which includes the modality start token)
+
+                offset += modality_length + precede_modality_tokens + succeed_modality_tokens # +2 due to [som] and [eom] - then account for meta start id and modality shape information (or eventually any meta information about modality)
 
-                batch_text.append(text_tensor)
                 modality_tensor = rearrange(modality_tensor, '... d -> (...) d')
 
                 batch_modality_tokens.append(modality_tensor)
-
-                batch_modality_positions.append((modality_type, offset + prec_meta_tag_length, modality_length)) # offset + preceding meta tag length (which includes the modality start token)
-
-                offset += modality_length + 2 + len(modality_meta_info) + 1 # +2 due to [som] and [eom] - then account for meta start id and modality shape information (or eventually any meta information about modality)
+                batch_text.append(text_tensor)
 
                 # handle axial positional embedding
 
@@ -1875,7 +1889,7 @@ class Transfusion(Module):
                     if exists(modality_pos_mlp):
                         pos_emb = modality_pos_mlp(tensor(modality_shape_tuple), flatten_dims= True)
 
-                        pos_emb = F.pad(pos_emb, (0, 0, prec_meta_tag_length, 1), value = 0.)
+                        pos_emb = F.pad(pos_emb, (0, 0, precede_modality_tokens, succeed_modality_tokens), value = 0.)
                     else:
                         pos_emb = torch.zeros(text_tensor.shape[0], self.dim, device = device)
 
@@ -2109,7 +2123,7 @@ class Transfusion(Module):
         # only the token positions that are not modalities have autoregressive loss
 
         total_loss = (
-            text_loss * text_loss_weight +
+            text_loss * text_loss_weight * self.text_loss_weight +
             (torch.stack(flow_losses) * modality_loss_weights).sum() * self.flow_loss_weight
         )
 
